"""
Base class for all Hoppr plug-ins
"""
from __future__ import annotations

import functools
import subprocess
import time
import traceback

from abc import ABC, abstractmethod
from os import PathLike
from typing import Any, Callable, Dict, List, Optional

import hoppr.plugin_utils
import hoppr.utils

from hoppr.logger import HopprLogger
from hoppr.models import HopprContext
from hoppr.models.transfer import ComponentCoverage
from hoppr.models.types import BomAccess
from hoppr.result import Result


def _get_component(*args, **kwargs) -> Optional[Any]:
    # TODO: This will only utilize the first Component found in 1) args, 2) kwargs  # pylint: disable=fixme
    # and assumes all plugins operate this way
    comp = next((arg for arg in args if type(arg).__name__ == "Component"), None)

    if comp is None:
        for value in kwargs.values():
            if type(value).__name__ == "Component":
                comp = value
                break

    return comp


def hoppr_process(func: Callable) -> Callable:
    """
    Decorator to handle generic bookkeeping for hoppr plug-ins
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs) -> Result:
        self._start_time = time.time()  # pylint: disable=protected-access

        comp = _get_component(*args, **kwargs)

        arg_string = ""
        if comp is not None:
            arg_string = f"(purl: {comp.purl})"

        self.get_logger().debug(f"Starting {self.__class__.__name__}.{func.__name__} {arg_string}")

        result = None
        if comp is not None:
            if comp.purl is None:
                result = Result.fail("No purl supplied for component")

            else:
                purl_type = hoppr.utils.get_package_url(comp.purl).type

                if not self.supports_purl_type(purl_type):
                    return Result.skip(f"Class {self.__class__.__name__} does not support purl type {purl_type}")

                component_header = f"{comp.name}" + (f"@{comp.version}" if comp.version is not None else "")

                self.get_logger().info(
                    msg=f"{'-' * 4} Component: {component_header} {'-' * 50}",
                )

        if result is None:
            # Only check for missing commands if func has been overridden
            if func.__module__ != HopprPlugin.__module__:
                command_result = hoppr.plugin_utils.check_for_missing_commands(self.required_commands)
                if command_result.is_fail():
                    self.get_logger().error(command_result.message)
                    return command_result

            try:
                result = func(self, *args, **kwargs)
            except Exception as error:  # pylint: disable=broad-except
                self.get_logger().error(
                    f"Unexpected exception running {self.__class__.__name__}.{func.__name__}: {error}",
                )
                self.get_logger().error(traceback.format_exc())
                result = Result.fail(f"Unexpected Exception: {error}")

        self.get_logger().debug(f"Completed {self.__class__.__name__}.{func.__name__}")

        duration = time.time() - self._start_time  # pylint: disable=protected-access
        self.get_logger().debug(f"Process duration {duration:3f} seconds")
        self.get_logger().info(f"Result: '{result}'")

        if result.is_skip():
            self.get_logger().clear_targets()

        self.close_logger()

        return result

    return wrapper


def hoppr_rerunner(method: Callable) -> Callable:
    """
    Runs a method (assumed to return a Result object) a number of times,
    or until the result type is not RETRY
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs) -> Result:
        log = self.get_logger()

        result = Result.fail(message="Context max_attempts must be a positive integer.")

        for attempt in range(self.context.max_attempts):
            log.info(msg=f"Processing component [attempt {attempt + 1} of {self.context.max_attempts}]", indent_level=1)

            result = method(self, *args, **kwargs)

            if not isinstance(result, Result):
                msg = f"Method {method.__name__} returned {type(result).__name__} in rerunner. Result object required"

                log.error(msg=msg, indent_level=1)
                return Result.fail(message=msg)

            if not result.is_retry():
                return result

            if attempt < self.context.max_attempts - 1:
                log.warning(
                    msg=(f"Method {method.__name__} will be retried in {self.context.retry_wait_seconds} seconds"),
                    indent_level=1,
                )

                log.warning(msg=f"Result message for attempt {attempt + 1}: {result.message}", indent_level=1)

                time.sleep(self.context.retry_wait_seconds)

        log.error(msg=f"Method {method.__name__} failed after {self.context.max_attempts} attempts")
        log.error(msg=f"Result message for final attempt: {result.message}")

        return Result.fail(f"Failure after {self.context.max_attempts} attempts, final message {result.message}")

    return wrapper


def hoppr_ignore_excluded(func: Callable) -> Callable:
    """
    Decorator to ignore components which have a scope of `excluded`
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs) -> Result:
        comp = _get_component(*args, **kwargs)
        if comp is not None and str(comp.scope) in {
            "excluded",
            "Scope.excluded",
        }:
            return Result.excluded()
        return func(self, *args, **kwargs)

    return wrapper


class HopprPlugin(ABC):
    """
    Base class for all Hoppr plug-ins

    Note that this class is not thread safe
    """

    required_commands: List[str] = []
    supported_purl_types: List[str] = []

    """
    Product: the result of carrying out a step. Products are usually persistent (e.g,.
    files), and are often meant to be used as materials on subsequent steps. Products are
    recorded as part of link metadata.

    This is a list of strings where each element is a relative path file or wildcard.

    See in-toto specification:
    - https://github.com/in-toto/docs/blob/master/in-toto-spec.md
    """
    products: List[str] = []

    default_component_coverage = ComponentCoverage.OPTIONAL
    bom_access = BomAccess.NO_ACCESS
    process_timeout = 60

    observers: dict[object, Callable] = {}

    def __init__(
        self,
        context: HopprContext,
        config: Optional[Dict] = None,
    ) -> None:
        self._start_time: float = 0.0
        self.config = config
        self.context = context

        self._logger = HopprLogger(
            filename=str(self.context.logfile_location),
            level=self.context.log_level,
            lock=self.context.logfile_lock,
        )

        self.process_timeout = (self.config or {}).get("process_timeout", self.process_timeout)

    @abstractmethod
    def get_version(self) -> str:
        """
        Returns the version of this plug-in
        """

    def notify(self, *args, **kwargs) -> None:
        """
        Call the callback function for all registered subscribers
        """
        for callback in self.observers.values():
            callback(*args, **kwargs)

    def subscribe(self, observer: object, callback: Callable) -> None:
        """
        Register an observer
        """
        self.observers[observer] = callback

    def unsubscribe(self, observer: object) -> None:
        """
        Unregister an observer
        """
        self.observers.pop(observer, None)

    @hoppr_process
    @hoppr_rerunner
    def pre_stage_process(self) -> Result:
        """
        Process to be run before other processing within a stage for this plug-in
        """

        return Result.skip("pre_stage_process not defined.")

    @hoppr_process
    @hoppr_rerunner
    def process_component(self, comp: Any) -> Result:  # pylint: disable=unused-argument
        """
        Process a single component through this plug-in
        """

        return Result.skip("process_component not defined.")

    @hoppr_process
    @hoppr_rerunner
    def post_stage_process(self):
        """
        Finalize processing for this plug-in
        """
        return Result.skip("post_stage_process not defined.")

    def supports_purl_type(self, purl_type: str) -> bool:
        """
        Indicates whether or not this particular plug-in supports components of the specified
        purl type.

        If no supported purl types are defined, the plug-in supports all purl types.
        """
        return purl_type in self.supported_purl_types or not self.supported_purl_types

    @classmethod
    def get_attestation_products(cls, config: dict | None = None) -> list[str]:  # pylint: disable=unused-argument
        """
        Return a list of attestation products for this class
        """
        return cls.products

    def get_logger(self) -> HopprLogger:
        """
        Returns the logger to be used for the current process
        """
        return self._logger

    def close_logger(self) -> None:
        """
        Close (and flush) all handlers for this plug-in's logger
        """
        self._logger.close()

    def run_command(
        self, command: list[str], password_list: list[str] | None = None, cwd: str | PathLike[str] | None = None
    ) -> subprocess.CompletedProcess[bytes]:
        """
        Run a command and log any errors
        """
        obscured_command = hoppr.utils.obscure_passwords(command, password_list)
        self.get_logger().debug(msg=f"Running command: '{obscured_command}'", indent_level=2)

        try:
            result = subprocess.run(
                command, check=False, shell=False, capture_output=True, cwd=cwd, timeout=self.process_timeout
            )
        except subprocess.TimeoutExpired as timeout_expired:
            self.get_logger().error(
                'Command "%s" timed out after %d seconds', command[0], timeout_expired.timeout, indent_level=2
            )
            result = subprocess.CompletedProcess(
                command, returncode=124, stdout=timeout_expired.stdout or b"", stderr=timeout_expired.stderr or b""
            )

        if result.returncode != 0:
            # Prefix each line of the command's stdout and stderr with 4 spaces
            stdout_indented, stderr_indented = map(
                lambda output: "\n".join([f"    {line}" for line in output.decode("utf-8").split("\n")]),
                [result.stdout, result.stderr],
            )

            self.get_logger().debug(msg=f"{command[0]} command stdout content:\n{stdout_indented}", indent_level=2)
            self.get_logger().error(msg=f"{command[0]} command failed with error:\n{stderr_indented}", indent_level=2)

        return result



"""
Collector plugin for Golang images
"""
from __future__ import annotations
from pydantic import SecretStr

import requests

from requests import HTTPError
from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectGolangPlugin(SerialCollectorPlugin):
    """
    Collector plugin for Golang images
    """

    supported_purl_types: list[str] = ["golang"]
    products: list[str] = ["golang/*"]
    system_repositories: list[str] = ["https://proxy.golang.org"]

    def get_version(self) -> str:
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)
        self.get_logger().info(msg=f"Copying Golang package from {comp.purl}", indent_level=2)

        mod_url = RepositoryUrl(url=repo_url) / (purl.namespace or "") / purl.name / "@v" / f"{purl.version}.zip"

        authentication = None
        if creds is not None and isinstance(creds.password, SecretStr):
            authentication = HTTPBasicAuth(creds.username, creds.password.get_secret_value())

        package_name = f"{'_'.join(filter(None, [purl.name, purl.version]))}.zip"
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / package_name

        response = requests.get(
            mod_url.url, auth=authentication, timeout=120, proxies=self._get_proxies_for(repo_url), stream=True
        )

        try:
            response.raise_for_status()
        except HTTPError:
            msg = f"Failed to locate Golang package for {comp.purl}, return_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        with package_out_path.open(mode="ab") as out_file:
            for chunk in response.iter_content(chunk_size=None):
                out_file.write(chunk)

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)



"""
Plugin to create tar file bundle
"""
from __future__ import annotations

import re
import tarfile
import time

from pathlib import Path

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.result import Result


class TarBundlePlugin(HopprPlugin):
    """
    Plugin to create tar file bundle

    This plug-in supports the following config values:

    - compression: Specifies the compression to be applied to the output tar file. Defaults
        to "gz". Recognized values are "none", "gzip", "gz", "bzip2", "bz", "lmza", "xz"
    - tarfile_name: Name of the tar file to be created. Defaults to "bundle.tar.<comp>", where <comp>
        is the short form of the compression used. For "none" compression, default is "bundle.tar"
    """

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def post_stage_process(self) -> Result:
        """
        Tar-up the context.collect_root_dir directory
        """
        compression: str = (self.config or {}).get("compression", "gz").lower()

        match compression:
            case "none":
                compression = ""
            case "gzip" | "gz":
                compression = "gz"
            case "bzip2" | "bz2" | "bz":
                compression = "bz2"
            case "lzma" | "xz":
                compression = "xz"
            case _:
                return Result.fail(f"Unrecognized compression in config: {compression}")

        default_filename = Path.cwd() / f"bundle.tar{f'.{compression}' if compression else ''}"
        tarfile_name = Path((self.config or {}).get("tarfile_name", default_filename)).expanduser()

        if tarfile_name.exists():
            timestr = time.strftime("%Y%m%d-%H%M%S")
            base_name, suffix = filter(None, re.split(pattern=r"^(.*)(\.tar\.?.*)$", string=tarfile_name.name))
            tarfile_name = tarfile_name.with_name(f"{base_name}-{timestr}{suffix}")

        msg = f"Bundling collected artifacts into {tarfile_name}"
        self.get_logger().info(msg)
        self.notify(msg, type(self).__name__)

        try:
            with tarfile.open(tarfile_name, f"x:{compression}") as tar:
                tar.add(self.context.collect_root_dir, ".")
        except tarfile.ReadError as err:
            return Result.fail(f"Unable to create tarfile {tarfile_name}, {err}")
        except FileNotFoundError:
            return Result.fail(f"File {tarfile_name}: Directory not found.")
        except PermissionError:
            return Result.fail(f"File {tarfile_name}: Permission denied.")

        return Result.success()



"""
Generate JSON schema files from pydantic models
"""
from __future__ import annotations

from pathlib import Path

from pydantic_yaml.compat.yaml_lib import yaml_safe_dump

from hoppr.models import CredentialsFile, HopprBaseModel, HopprSchemaModel, ManifestFile, TransferFile


def write_json_schema(filename: str, model_type: type[HopprBaseModel]) -> None:
    """
    Write JSON schema to file
    """
    with Path(filename).open(mode="w+", encoding="utf-8") as json_schema:
        json_schema.write(model_type.schema_json(indent=2))


def write_yaml_schema(filename: str, model_type: type[HopprBaseModel]) -> None:
    """
    Write YAML schema to file
    """
    with Path(filename).open(mode="w+", encoding="utf-8") as yaml_schema:
        yaml_safe_dump(data=model_type.schema(by_alias=True), stream=yaml_schema, default_flow_style=False)


# Write JSON schema files
write_json_schema(filename="hoppr-credentials-schema-v1.json", model_type=CredentialsFile)
write_json_schema(filename="hoppr-manifest-schema-v1.json", model_type=ManifestFile)
write_json_schema(filename="hoppr-transfer-schema-v1.json", model_type=TransferFile)
write_json_schema(filename="hoppr-combined-schema-v1.json", model_type=HopprSchemaModel)

# Write YAML schema files
write_yaml_schema(filename="hoppr-credentials-schema-v1.yml", model_type=CredentialsFile)
write_yaml_schema(filename="hoppr-manifest-schema-v1.yml", model_type=ManifestFile)
write_yaml_schema(filename="hoppr-transfer-schema-v1.yml", model_type=TransferFile)
write_yaml_schema(filename="hoppr-combined-schema-v1.yml", model_type=HopprSchemaModel)



"""
Report Generating Plugin
"""

import shutil

from pathlib import Path
from typing import Any, Optional
from uuid import UUID

import typer

from jinja2 import Environment, FileSystemLoader

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin
from hoppr.models.manifest import SearchSequence
from hoppr.models.sbom import Component
from hoppr.result import ResultStatus


class Report:  # pylint: disable=too-few-public-methods
    """
    A report of a plugin's execution stage result.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        unique_id: UUID,
        plugin: str,
        stage: str,
        result: ResultStatus,
        details: Any,
        method: str,
        component: Optional[Component] = None,
    ):
        self.unique_id = unique_id
        self.plugin = plugin
        self.stage = stage
        self.result = result
        self.details = details
        self.method = method
        self.component = component


class ReportGenerator(HopprPlugin):
    """
    A class for generating a report based on a list of `Report` objects.
    """

    report_gen_list: list[Report] = []

    def get_version(self) -> str:
        return __version__

    def generate_report(self) -> None:  # pylint: disable=too-many-locals
        """
        Generate a report based on the given list of `Report` objects.
        """

        # Count the number of reports and errors
        report_count = len(self.report_gen_list)
        error_count = len([report for report in self.report_gen_list if report.result.numerator == ResultStatus.FAIL])

        # Sort reports
        reports_with_components = [report for report in self.report_gen_list if report.component is not None]
        sorted_reports = sorted(reports_with_components, key=lambda report: (-report.result.numerator, report.plugin))

        # Organize reports by plugin and stage
        reports_by_plugin = self._get_reports_by("plugin", sorted_reports)
        reports_by_stage = self._get_reports_by("stage", sorted_reports)

        # Generate report HTML file
        environment = Environment(loader=FileSystemLoader(Path(__file__).parent / "templates"))
        index_template = environment.get_template("index.jinja2")

        report_dir = Path.cwd() / "report"
        report_dir.mkdir(exist_ok=True)

        results_file_name = report_dir / "index.html"

        for dir_name in ("assets", "scripts", "styles"):
            source_dir = Path(__file__).parent / dir_name
            dest_dir = report_dir / dir_name
            shutil.copytree(src=source_dir, dst=dest_dir, dirs_exist_ok=True)

        context = {
            "report_count": report_count,
            "error_count": error_count,
            "reports_by_stage": reports_by_stage,
            "reports_by_plugin": reports_by_plugin,
            "ResultStatus": ResultStatus,
            "get_type": type,
            "get_length": len,
            "parse_search_sequence": SearchSequence.parse_raw,
            "get_successful_report_total": self._get_successful_report_total,
            "get_overall_status_button_class": self._get_overall_status_button_class,
        }

        results_file_name.write_text(index_template.render(context), encoding="utf-8")
        typer.echo(f"HTML report generated: {results_file_name}")

    def _get_successful_report_total(self, reports: list[Report]) -> int:
        """
        Return the number of successful reports for a given list of reports.
        """

        successful_reports = [report for report in reports if report.result.numerator == ResultStatus.SUCCESS]
        return len(successful_reports)

    def _get_overall_status_button_class(self, reports: list[Report]) -> str:
        """
        Determine the overall status of a list of reports and returns the appropriate css class.
        """

        num_successful_reports = sum(report.result.numerator == ResultStatus.SUCCESS for report in reports)

        if num_successful_reports == 0:
            return "btn-danger"

        if num_successful_reports == len(reports):
            return "btn-success"

        return "btn-warning"

    def _get_reports_by(self, sort_by: str, reports: list[Report]) -> dict:
        """
        Organize report data by plugin or stage.
        """

        sorted_reports: dict = {}

        for report in reports:
            if report.__dict__[sort_by] in sorted_reports:
                sorted_reports[report.__dict__[sort_by]].append(report)
            else:
                sorted_reports[report.__dict__[sort_by]] = [report]

        return sorted_reports



"""
`validate` subcommand for `hoptctl`
"""
from __future__ import annotations

import functools
import json
import logging
import sys

from datetime import date, datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Literal, Mapping, TypeAlias

import typer

from rich.console import Group
from rich.live import Live
from rich.progress import Progress
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from typer import Argument, CallbackParam, Context, Option, Typer

import hoppr.main
import hoppr.utils

from hoppr.cli.layout import HopprBasePanel, HopprJobsPanel, HopprLayout, HopprSpinnerColumn, console
from hoppr.cli.options import basic_term_option, log_file_option, sbom_dirs_option, sbom_files_option, verbose_option
from hoppr.exceptions import HopprLoadDataError
from hoppr.logger import HopprLogger
from hoppr.models.licenses import LicenseExpressionItem, LicenseMultipleItem
from hoppr.models.sbom import Component, Metadata, Sbom
from hoppr.models.validate import (
    ValidateCheckResult,
    ValidateComponentResult,
    ValidateLicenseResult,
    ValidateOutputFormat,
    ValidateSbomResult,
)
from hoppr.result import ResultStatus

if TYPE_CHECKING:
    from rich.console import Console, ConsoleOptions, RenderResult

    DictStrAny: TypeAlias = dict[str, Any]

CURRENT_DATE: Final[date] = date.today()
EXPIRATION_DAYS: float = 30

RESULT_MAP: Final[Mapping[ResultStatus, Text]] = {
    ResultStatus.FAIL: Text("\u274c\ufe0f"),
    ResultStatus.SUCCESS: Text.from_markup("[green]\u2714"),
    ResultStatus.WARN: Text("\u26a0\ufe0f"),
}

STRICT_ALL = False
STRICT_LICENSE = False
STRICT_NTIA = False

layout: HopprValidateLayout
logger: HopprLogger

app = Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Validate CycloneDX SBOMs or Hoppr config files",
    no_args_is_help=True,
    rich_markup_mode="markdown",
)

LICENSE_FAIL_LOG_LEVEL: int
LICENSE_FAIL_RESULT: ValidateCheckResult
NTIA_FAIL_LOG_LEVEL: int
NTIA_FAIL_RESULT: ValidateCheckResult


class HopprValidateJobsPanel(HopprJobsPanel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress("{task.description}", HopprSpinnerColumn(), expand=True)

    def __init__(self) -> None:
        super().__init__()
        self.renderable = self.progress_bar
        self.title = "[bold blue]SBOM Files"


class HopprValidateLayout(HopprLayout):
    """
    Layout of the `hopctl validate sbom` console application
    """

    name: str = "root"
    jobs_panel = HopprValidateJobsPanel()


class HopprValidateSummary(HopprBasePanel):  # pylint: disable=too-few-public-methods
    """
    `hopctl validate sbom` results summary
    """

    results: list[ValidateSbomResult] = []
    summary_group = Group()

    def __init__(self, *args, title: str | None = "Summary", log_file: Path, **kwargs) -> None:
        self.log_file = log_file

        super().__init__(self.summary_group, title, *args, **kwargs)

    def _component_ntia_row(self, component_results: list[ValidateComponentResult]) -> tuple[str, Text]:
        component_ntia_fails = [cr for cr in component_results if not cr.ntia_fields_result.is_success()]
        ntia_fail_status = RESULT_MAP[NTIA_FAIL_RESULT.status]

        return (
            f"{len(component_ntia_fails)} out of {len(component_results)} components missing minimum NTIA fields",
            ntia_fail_status if component_ntia_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def _license_fields_row(
        self,
        license_results: list[ValidateLicenseResult],
        result_type: Literal["component", "SBOM"],
    ) -> tuple[str, Text]:
        license_field_fails = [lr for lr in license_results if not lr.required_fields.is_success()]

        license_fail_status = RESULT_MAP[LICENSE_FAIL_RESULT.status]

        return (
            f"{len(license_field_fails)} out of {len(license_results)} "
            f"{result_type} licenses missing minimum license fields",
            license_fail_status if license_field_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def _license_expiration_row(
        self,
        license_results: list[ValidateLicenseResult],
        result_type: Literal["component", "SBOM"],
    ) -> tuple[str, Text]:
        license_exp_fails = [lr for lr in license_results if not lr.expiration.is_success()]

        license_fail_status = RESULT_MAP[LICENSE_FAIL_RESULT.status]

        return (
            f"{len(license_exp_fails)} out of {len(license_results)} "
            f"{result_type} licenses expired or expiring within {EXPIRATION_DAYS:n} days",
            license_fail_status if license_exp_fails else RESULT_MAP[ResultStatus.SUCCESS],
        )

    def __rich_console__(  # pylint: disable=arguments-renamed
        self, console_: Console, options: ConsoleOptions
    ) -> RenderResult:
        for sbom_result in self.results:
            component_license_checks = [lr for cr in sbom_result.component_results for lr in cr.license_results]

            self.summary_group.renderables.append(f"Results for {sbom_result.sbom_file}:")
            summary_table = Table.grid(padding=(0, 4), pad_edge=True)

            for result in [sbom_result.spec_version_result, sbom_result.ntia_fields_result]:
                summary_table.add_row(result.message, RESULT_MAP[result.status])

            # Add SBOM metadata license results summary rows
            summary_table.add_row(*self._license_fields_row(sbom_result.license_results, "SBOM"))
            summary_table.add_row(*self._license_expiration_row(sbom_result.license_results, "SBOM"))

            # Add component NTIA minimum fields and license results summary rows
            summary_table.add_row(*self._component_ntia_row(sbom_result.component_results))
            summary_table.add_row(*self._license_fields_row(component_license_checks, "component"))
            summary_table.add_row(*self._license_expiration_row(component_license_checks, "component"))

            self.summary_group.renderables.append(summary_table)
            self.summary_group.renderables.append(Rule(style="[white]rule.line"))

        self.summary_group.renderables.append(f"See log file {self.log_file} for full results.")

        if hoppr.utils.is_basic_terminal():
            console_.line()
            console_.rule(title="Summary", characters="=", style="[bold]rule.line")
            console_.print(self.summary_group)
            return []

        return super().__rich_console__(console_, options)


def _init_globals(  # pylint: disable=global-statement, too-many-arguments
    log_file: Path,
    verbose: bool,
    expiration_days: float,
    strict: bool,
    strict_license_fields: bool,
    strict_ntia_minimum_fields: bool,
):
    global layout, logger

    logger = HopprLogger(
        name="hopctl-validate",
        filename=str(log_file),
        level=logging.DEBUG if verbose else logging.INFO,
    )

    layout = HopprValidateLayout()

    global EXPIRATION_DAYS, STRICT_ALL, STRICT_LICENSE, STRICT_NTIA

    EXPIRATION_DAYS = expiration_days
    STRICT_ALL = strict
    STRICT_LICENSE = strict_license_fields
    STRICT_NTIA = strict_ntia_minimum_fields

    global LICENSE_FAIL_LOG_LEVEL, LICENSE_FAIL_RESULT, NTIA_FAIL_LOG_LEVEL, NTIA_FAIL_RESULT

    LICENSE_FAIL_LOG_LEVEL = logging.ERROR if STRICT_LICENSE else logging.WARN
    LICENSE_FAIL_RESULT = ValidateCheckResult.fail() if STRICT_LICENSE else ValidateCheckResult.warn()
    NTIA_FAIL_LOG_LEVEL = logging.ERROR if STRICT_NTIA else logging.WARN
    NTIA_FAIL_RESULT = ValidateCheckResult.fail() if STRICT_NTIA else ValidateCheckResult.warn()


def _set_all_strict(ctx: Context, param: CallbackParam, value: bool) -> bool:
    """
    Set all strict flags to `True`

    Args:
        ctx (Context): Click context to update
        param (CallbackParam): Typer metadata for the parameter
        value (bool): The parameter value

    Returns:
        bool: The flag that was passed, unmodified
    """
    if param.name and value:
        ctx.params[param.name] = True
        ctx.params["strict_ntia_minimum_fields"] = True
        ctx.params["strict_license_fields"] = True

    return value


def _check_fields(
    obj: object | None,
    *field_names: str,
    fail_log_level: int = logging.WARN,
    fail_result: ValidateCheckResult = ValidateCheckResult.warn(),
    indent_level: int = 1,
) -> ValidateCheckResult:
    result = ValidateCheckResult.success()

    if isinstance(obj, dict):
        field_values = (obj.get(field_name) for field_name in field_names)
    else:
        field_values = (getattr(obj, field_name, None) for field_name in field_names)

    if obj is None or not any(field_values):
        missing_fields = ", ".join(f"`{name}`" for name in field_names)
        logger.log(
            fail_log_level,
            "Missing %s field",
            missing_fields,
            indent_level=indent_level,
        )

        result.merge(fail_result)

    return result


def _check_license_expiration(license_: LicenseMultipleItem | LicenseExpressionItem) -> ValidateCheckResult:
    if isinstance(license_, LicenseExpressionItem) or not license_.license.licensing:
        return LICENSE_FAIL_RESULT

    try:
        # Raises `ValueError` if `expiration` is `None`
        expiration = datetime.fromisoformat(str(license_.license.licensing.expiration))
        if expiration.date() < CURRENT_DATE + timedelta(days=EXPIRATION_DAYS):
            raise ValueError
    except ValueError:
        logger.log(
            LICENSE_FAIL_LOG_LEVEL,
            "License expired or expiring within %d days",
            EXPIRATION_DAYS,
            indent_level=3,
        )

        return LICENSE_FAIL_RESULT

    return ValidateCheckResult.success()


def _check_license_fields(license_: LicenseMultipleItem | LicenseExpressionItem) -> ValidateCheckResult:
    required_fields_result = ValidateCheckResult.success()

    if isinstance(license_, LicenseExpressionItem):
        required_fields_result.merge(LICENSE_FAIL_RESULT)
        return required_fields_result

    check_license_fields = functools.partial(
        _check_fields,
        fail_log_level=LICENSE_FAIL_LOG_LEVEL,
        fail_result=LICENSE_FAIL_RESULT,
        indent_level=3,
    )

    required_fields_result.merge(check_license_fields(license_.license, "id", "name"))
    required_fields_result.merge(check_license_fields(license_.license, "licensing"))
    return required_fields_result


def _check_spec_version(sbom_dict: dict[str, Any]) -> ValidateCheckResult:
    result = ValidateCheckResult.success()

    match sbom_dict.get("specVersion"):
        case "1.2" | None:
            result.merge(ValidateCheckResult.fail())
        case "1.3" | "1.4":
            result.merge(ValidateCheckResult.fail() if STRICT_ALL else ValidateCheckResult.warn())

    return result


def _validate_component(component: Component) -> ValidateComponentResult:
    component_id = "@".join(filter(None, [component.name, component.version]))
    comp_result = ValidateComponentResult(component_id=component_id)

    logger.info("Validating component: %s", component_id, indent_level=1)

    check_component_fields = functools.partial(
        _check_fields,
        fail_log_level=NTIA_FAIL_LOG_LEVEL,
        fail_result=NTIA_FAIL_RESULT,
        indent_level=2,
    )

    # Validate minimum required NTIA fields for component
    comp_result.ntia_fields_result.merge(check_component_fields(component, "supplier"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "name"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "version"))
    comp_result.ntia_fields_result.merge(check_component_fields(component, "cpe", "purl", "swid"))
    comp_result.result.merge(comp_result.ntia_fields_result)

    license_results = _validate_licenses(component)

    for license_result in license_results:
        comp_result.license_results.append(license_result)

        # Merge license result into final result for this SBOM
        comp_result.result.merge(license_result.required_fields)
        comp_result.result.merge(license_result.expiration)

    return comp_result


def _validate_licenses(obj: Metadata | Component | None) -> list[ValidateLicenseResult]:
    license_results: list[ValidateLicenseResult] = []

    if not obj or not obj.licenses:
        logger.log(LICENSE_FAIL_LOG_LEVEL, "Missing `licenses` field", indent_level=2)

        license_result = ValidateLicenseResult(license_id="Missing license data")
        license_result.required_fields.merge(LICENSE_FAIL_RESULT)
        license_result.expiration.merge(LICENSE_FAIL_RESULT)
        license_results.append(license_result)

        return license_results

    for license_ in obj.licenses:
        license_id = (
            license_.expression
            if isinstance(license_, LicenseExpressionItem)
            else str(license_.license.id or license_.license.name)
        )

        logger.info("Validating license: %s", license_id, indent_level=2)

        license_result = ValidateLicenseResult(license_id=license_id)
        license_result.required_fields.merge(_check_license_fields(license_))

        # Validate license expiration not within specified number of days
        license_result.expiration.merge(_check_license_expiration(license_))

        license_results.append(license_result)

    return license_results


def _validate_sbom(sbom_file: Path) -> ValidateSbomResult:
    sbom_result = ValidateSbomResult(sbom_file=sbom_file)

    try:
        sbom_dict = hoppr.utils.load_file(sbom_file)
        if not isinstance(sbom_dict, dict):
            raise TypeError("SBOM file was not loaded as dictionary")

        # Check raw `specVersion` and NTIA minimum fields values before parsing data as Sbom object
        sbom_result.spec_version_result.merge(_check_spec_version(sbom_dict))

        check_sbom_fields = functools.partial(
            _check_fields,
            fail_log_level=NTIA_FAIL_LOG_LEVEL,
            fail_result=NTIA_FAIL_RESULT,
            indent_level=1,
        )

        sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_dict, "metadata"))
        sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_dict.get("metadata"), "authors", "tools"))
        sbom_result.ntia_fields_result.merge(check_sbom_fields(sbom_dict.get("metadata"), "timestamp"))
        sbom_result.result.merge(sbom_result.ntia_fields_result)

        logger.info("Validating SBOM licenses", indent_level=1)

        # Parse SBOM data and run all validation checks
        sbom_ = Sbom.load(sbom_file)
        sbom_result.license_results = _validate_licenses(sbom_.metadata)

        # Merge license results into final result for this SBOM
        for license_result in sbom_result.license_results:
            sbom_result.result.merge(license_result.required_fields)
            sbom_result.result.merge(license_result.expiration)

        # Skip validation for components with type of `operating-system`
        for component in [comp for comp in sbom_.components if str(comp.type) != "operating-system"]:
            component_result = _validate_component(component)
            sbom_result.component_results.append(component_result)

            # Merge component NTIA fields result into final result for this SBOM
            sbom_result.ntia_fields_result.merge(component_result.ntia_fields_result)

            # Merge component result into final result for this SBOM
            sbom_result.result.merge(component_result.result)
    except (HopprLoadDataError, TypeError) as ex:
        sbom_result.result.merge(ValidateCheckResult.fail(message=str(ex)))

    return sbom_result


@app.command(no_args_is_help=True)
def sbom(  # pylint: disable=too-many-locals, too-many-arguments, unused-argument
    sbom_files: list[Path] = sbom_files_option,
    sbom_dirs: list[Path] = sbom_dirs_option,
    strict: bool = Option(
        False,
        "-S",
        "--strict",
        callback=_set_all_strict,
        help="Enable all strict validation options",
        show_default=False,
    ),
    strict_ntia_minimum_fields: bool = Option(
        False,
        "-n",
        "--strict-ntia-minimum-fields",
        help="Raise error if minimum fields recommended by NTIA are not set",
        is_eager=True,
        show_default=False,
    ),
    strict_license_fields: bool = Option(
        False,
        "-l",
        "--strict-license-fields",
        help="Raise error if SBOM license or SBOM/component license expiration fields are not set",
        is_eager=True,
        show_default=False,
    ),
    expiration_days: float = Option(
        30,
        "-d",
        "--expiration-days",
        help="Number of days allowed by license expiration check",
        show_default=False,
    ),
    output_format: ValidateOutputFormat = Option(
        ValidateOutputFormat.JSON,
        "-f",
        "--output-format",
        help=f"Format for output file {typer.style(text='[default: json]', dim=True)}",
        show_default=False,
    ),
    output_file: Path = Option(
        None,
        "-o",
        "--output-file",
        dir_okay=False,
        exists=False,
        help="Path to output file",
        resolve_path=True,
        show_default=False,
    ),
    basic_term: bool = basic_term_option,
    log_file: Path = log_file_option,
    verbose: bool = verbose_option,
):  # pragma: no cover
    """
    Validate SBOM file(s)
    """
    _init_globals(log_file, verbose, expiration_days, strict, strict_license_fields, strict_ntia_minimum_fields)

    live_display = Live(layout, console=console, refresh_per_second=10)

    summary_panel = HopprValidateSummary(log_file=log_file)

    result = ValidateCheckResult.success()

    sbom_files = hoppr.utils.dedup_list(sbom_files)

    for sbom_file in sbom_files:
        layout.add_job(description=sbom_file.name)

    if not hoppr.utils.is_basic_terminal():
        live_display.start(refresh=True)

    sbom_results: list[ValidateSbomResult] = []

    for sbom_file in sbom_files:
        msg = f"Validating {sbom_file}..."
        logger.info(msg)
        layout.print(msg)

        sbom_result = _validate_sbom(sbom_file)
        summary_panel.results.append(sbom_result)
        result.merge(sbom_result.result)
        sbom_results.append(sbom_result)

        layout.update_job(name=sbom_file.name, status=result.status.name)
        layout.stop_job(sbom_file.name)
        layout.advance_job(sbom_file.name)

    if not hoppr.utils.is_basic_terminal():
        live_display.stop()

    console.print("\n", summary_panel)

    if output_file:
        # Can be extended in the future to write other types of reports
        output_data = json.loads(f"[{', '.join([sbom_result.json() for sbom_result in sbom_results])}]")
        output_file.write_text(data=json.dumps(output_data, indent=2), encoding="utf-8")

    sys.exit(1 if result.is_fail() else 0)


@app.command(no_args_is_help=True)
def config(
    input_files: list[Path] = Argument(
        ...,
        dir_okay=False,
        exists=True,
        help="Path to manifest file",
        resolve_path=True,
        show_default=False,
    ),
    credentials_file: Path = Option(
        None,
        "-c",
        "--credentials",
        help="Specify credentials config for services",
        envvar="HOPPR_CREDS_CONFIG",
        show_default=False,
    ),
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
):  # pragma: no cover
    """
    Validate Hoppr manifest, transfer, and credentials file(s)
    """
    hoppr.main.validate(input_files, credentials_file, transfer_file)



"""
`generate` subcommand for `hoptctl`
"""
from __future__ import annotations

import os

from pathlib import Path
from runpy import run_module

import typer

from typer import Option, Typer

from hoppr import main

app = Typer(
    name="generate",
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Generate `in-toto` keys/layout or schemas for Hoppr input files",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
)


@app.command()
def layout(
    # pylint: disable=unused-argument
    transfer_file: Path = Option(
        "transfer.yml",
        "-t",
        "--transfer",
        envvar="HOPPR_TRANSFER_CONFIG",
        help="Specify transfer config",
    ),
    project_owner_key_path: Path = Option(
        ...,
        "-pk",
        "--project-owner-key",
        envvar="HOPPR_PROJECT_OWNER_KEY",
        help="Path to key used to sign in-toto layout",
    ),
    functionary_key_path: Path = Option(
        ...,
        "-fk",
        "--functionary-key",
        envvar="HOPPR_FUNCTIONARY_KEY",
        help="Path to key used to sign in-toto layout",
    ),
    project_owner_key_prompt: bool = Option(
        False,
        "-p",
        "--prompt",
        envvar="HOPPR_PROJECT_OWNER_KEY_PROMPT",
        help="Prompt user for project owner key's password",
    ),
    project_owner_key_password: str = Option(
        None,
        "-pk-pass",
        "--project-owner-key-password",
        envvar="HOPPR_PROJECT_OWNER_KEY_PASSWORD",
        help="Password for project owner key",
    ),
):  # pragma: no cover
    """
    Create in-toto layout based on transfer file
    """
    main.generate_layout(**locals())


@app.command(hidden=True)
def schemas(
    output_dir: Path = Option(
        Path(),
        "-o",
        "--output-dir",
        exists=True,
        file_okay=False,
        help=f"Output directory for schema files {typer.style(text='[default: current directory]', dim=True)}",
        resolve_path=True,
        show_default=False,
    )
):  # pragma: no cover
    """
    Generate JSON/YAML schemas for Hoppr manifest, credential, and transfer files
    """
    os.chdir(output_dir)
    run_module(mod_name="hoppr.models")



"""
Plugin to remove SBOM components that are specified by a "previous" SBOM
"""
from __future__ import annotations

import io
import tarfile

from pathlib import Path
from typing import TYPE_CHECKING

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.exceptions import HopprError
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import Component, Sbom
from hoppr.models.types import BomAccess
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL


class DeltaSbom(HopprPlugin):
    """
    Plugin to remove SBOM components that are specified by a "previous" SBOM
    """

    bom_access = BomAccess.FULL_ACCESS
    products: list[str] = ["generic/_metadata_/_previous_bom.json"]

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def pre_stage_process(self):
        """
        Tar-up the context.collect_root_dir directory
        """

        self.config = self.config or {}

        try:
            previous_source = str(next(filter(None, [self.context.previous_delivery, self.config.get("previous")])))
        except StopIteration:
            return Result.success(
                "No previously delivered bundle specified for delta bundle. All components will be delivered"
            )

        fail_on_empty = bool(self.config.get("fail_on_empty"))

        if not Path(previous_source).exists():
            return Result.fail(f'Previous source file "{previous_source}" not found.')

        self.get_logger().info("Creating delta/update SBOM, previous SBOM being retrieved from %s", previous_source)

        previous_sbom = self._get_previous_bom(previous_source)

        target_dir = self.context.collect_root_dir / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        with (target_dir / "_previous_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(previous_sbom.json(exclude_none=True, by_alias=True, indent=2))

        delta_sbom = self.context.delivered_sbom.copy(deep=True)
        delta_sbom.components.clear()

        for new_comp in self.context.delivered_sbom.components or []:
            include_component = not any(
                DeltaSbom._component_match(new_comp, prev_comp) for prev_comp in previous_sbom.components or []
            )
            if include_component:
                self.get_logger().debug("Including purl %s", new_comp.purl, indent_level=1)
                delta_sbom.components.append(new_comp)

        self.get_logger().info("Input sbom has %d components", len(self.context.delivered_sbom.components))
        self.get_logger().info("Prev  sbom has %d components", len(previous_sbom.components))
        self.get_logger().info("Delta sbom has %d components", len(delta_sbom.components))

        if len(delta_sbom.components) == 0:
            msg = f'No components updated since "{previous_source}".'

            return Result.fail(msg) if fail_on_empty else Result.success(msg)

        return Result.success(
            f"Delivering updates for {len(delta_sbom.components)} of "
            f"{len(self.context.delivered_sbom.components)} components.",
            return_obj=delta_sbom,
        )

    @staticmethod
    def _get_previous_bom(source: str) -> Sbom:
        if tarfile.is_tarfile(name=source):
            with tarfile.open(source) as tar:
                buffer = tar.extractfile("./generic/_metadata_/_consolidated_bom.json")

                if buffer is None:
                    raise HopprError("Unable to extract BOM file from tar")

                with io.TextIOWrapper(buffer) as bom_file:
                    content: str = bom_file.read()
                    bom_dict = hoppr.utils.load_string(content)

                    if not isinstance(bom_dict, dict):
                        raise HopprError("Invalid BOM file retrieved from tar")

                    return Sbom(**bom_dict)

        Sbom.loaded_sboms.clear()
        return Manifest.load(Path(source)).consolidated_sbom

    @staticmethod
    def _purl_match(new_purl: PackageURL, prev_purl: PackageURL) -> bool:
        if any(
            getattr(new_purl, attr) != getattr(prev_purl, attr)
            for attr in ["name", "type", "namespace", "version", "subpath"]
        ):
            return False

        qual_keys = list(new_purl.qualifiers.keys())
        qual_keys.extend(list(prev_purl.qualifiers.keys()))

        return all(
            new_purl.qualifiers.get(key) == prev_purl.qualifiers.get(key) for key in hoppr.utils.dedup_list(qual_keys)
        )

    @staticmethod
    def _component_match(new_comp: Component, prev_comp: Component) -> bool:
        if not (new_comp.purl and prev_comp.purl):
            return False

        new_purl = hoppr.utils.get_package_url(new_comp.purl)
        prev_purl = hoppr.utils.get_package_url(prev_comp.purl)

        if not DeltaSbom._purl_match(new_purl, prev_purl):
            return False

        hash_matches = 0

        for new_hash in new_comp.hashes:
            for prev_hash in prev_comp.hashes:
                if new_hash.alg != prev_hash.alg:
                    continue
                if new_hash.content != prev_hash.content:
                    return False
                hash_matches += 1

        if hash_matches > 0:
            return True

        return new_purl.version is not None and new_purl.version != "latest"



"""
Collector plugin for maven artifacts
"""
from __future__ import annotations

import os
import warnings

from copy import deepcopy
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, OrderedDict

import jmespath
import xmltodict

from pydantic import SecretStr

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprExperimentalWarning
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL

_MAVEN_DEP_PLUGIN = "org.apache.maven.plugins:maven-dependency-plugin:3.5.0"


class _CollectMavenBase(SerialCollectorPlugin):
    supported_purl_types = ["maven"]
    required_commands = []
    products: list[str] = ["maven/*"]
    system_repositories: list[str] = ["https://repo.maven.apache.org/maven2"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.extra_opts: list[str] = []

        if self.config is not None:
            self.required_commands = self.config.get("maven_command", self.required_commands)
            self.extra_opts = self.config.get("maven_opts", self.extra_opts)

        system_settings_file = Path("/") / "etc" / "maven" / "settings.xml"
        user_settings_file = Path.home() / ".m2" / "settings.xml"

        if not self.context.strict_repos:
            # Identify system repositories
            for settings_file in [system_settings_file, user_settings_file]:
                if settings_file.is_file():
                    settings_dict: OrderedDict[str, Any] = xmltodict.parse(
                        settings_file.read_text(encoding="utf-8"),
                        encoding="utf-8",
                        force_list={"profile", "repository"},
                    )

                    repo_urls: list[str] = jmespath.search(
                        expression="settings.profiles.profile[].repositories.repository[].url", data=settings_dict
                    )

                    for repo in repo_urls or []:
                        if repo not in self.system_repositories:
                            self.system_repositories.append(repo)


if os.getenv("HOPPR_EXPERIMENTAL"):
    warnings.warn(
        message="This Maven collector plugin is experimental; use at your own risk.",
        category=HopprExperimentalWarning,
    )

    class CollectMavenPlugin(_CollectMavenBase):
        """
        Collector plugin for maven artifacts (EXPERIMENTAL)
        """

        def _check_artifact_hash(
            self,
            download_url: str,
            dest_file: str,
            creds: CredentialRequiredService | None = None,
        ) -> Result:
            with NamedTemporaryFile(mode="w+", encoding="utf-8") as hash_file:
                # Download SHA1 hash file for artifact
                response = hoppr.net.download_file(
                    url=f"{download_url}.sha1",
                    dest=hash_file.name,
                    creds=creds,
                    proxies=self._get_proxies_for(download_url),
                )

                result = Result.from_http_response(response=response)
                hash_string = Path(hash_file.name).read_text(encoding="utf-8").strip().lower()

            if (computed_hash := hoppr.net.get_file_hash(artifact=dest_file)) != hash_string:
                result.merge(Result.fail(message=f"SHA1 hash for {Path(dest_file).name} does not match expected hash."))
                self.get_logger().debug("Computed hash: %s, expected: %s", computed_hash, hash_string, indent_level=2)

            return result

        def _get_maven_component(
            self,
            purl: PackageURL,
            repo_url: str,
            target_dir: Path,
            creds: CredentialRequiredService | None = None,
        ) -> Result:
            artifact_path = "/".join((purl.namespace or "").split("."))
            artifact_type = purl.qualifiers.get("type", "jar")
            artifact_classifier = purl.qualifiers.get("classifier")
            artifact_file = "-".join(
                filter(None, [purl.name, purl.version, "" if artifact_type == "pom" else artifact_classifier])
            )
            artifact_file = f"{artifact_file}.{artifact_type}"

            download_url = str(RepositoryUrl(url=repo_url) / artifact_path / purl.name / purl.version / artifact_file)

            # Reconstructs the file name to match what is expected by downstream plugins
            dest_file = f"{target_dir / '_'.join(filter(None, [purl.name, purl.version]))}.{artifact_type}"

            self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
            self.get_logger().info(msg=f"destination: {dest_file}", indent_level=3)

            response = hoppr.net.download_file(
                url=download_url,
                dest=dest_file,
                creds=creds,
                proxies=self._get_proxies_for(download_url),
            )

            if not (result := Result.from_http_response(response=response)).is_success():
                return result

            result.merge(self._check_artifact_hash(download_url, dest_file, creds))

            return result

        @hoppr_rerunner
        def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
            purl = hoppr.utils.get_package_url(comp.purl)
            target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

            self.get_logger().info(msg="Downloading Maven artifact:", indent_level=2)
            result = self._get_maven_component(purl, repo_url, target_dir, creds)

            if not result.is_success():
                msg = f"Failed to download Maven artifact {comp.purl}, {result.message}"
                self.get_logger().error(msg=msg, indent_level=2)
                return Result.fail(message=msg)

            if purl.qualifiers.get("type") != "pom":
                self.get_logger().info(msg="Downloading pom for Maven artifact:", indent_level=2)

                purl_copy = deepcopy(purl)
                purl_copy.qualifiers["type"] = "pom"
                result = self._get_maven_component(purl_copy, repo_url, target_dir, creds)

                if not result.is_success():
                    msg = f"Failed to download pom for Maven artifact {comp.purl}, {result.message}"
                    self.get_logger().error(msg=msg, indent_level=2)
                    return Result.fail(message=msg)

            self.set_collection_params(comp, repo_url, target_dir)

            return Result.success(return_obj=comp)

else:

    class CollectMavenPlugin(_CollectMavenBase):  # type: ignore[no-redef]
        """
        Collector plugin for maven artifacts
        """

        required_commands = ["mvn"]

        def _get_maven_component(self, command: list[str], password_list: list[str], **kwargs):
            full_command = command.copy()
            full_command.extend(f"-D{key}={value}" for key, value in kwargs.items())
            result = self.run_command(full_command, password_list)

            # The maven plugin does not recognize the 'destFileName' argument, so rename the file
            if result.returncode == 0:
                _, name, version, extension = kwargs.get("artifact", "::::").split(":")
                directory = kwargs.get("outputDirectory", ".")
                old_name = f"{name}-{version}.{extension}"
                new_name = kwargs.get("destFileName", f"{name}_{version}.{extension}")

                os.rename(directory / old_name, directory / new_name)

            return result

        @hoppr_rerunner
        def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
            """
            Copy a component to the local collection directory structure
            """
            # pylint: disable=too-many-locals

            purl = hoppr.utils.get_package_url(comp.purl)
            artifact = f"{purl.namespace}:{purl.name}:{purl.version}"

            extension = purl.qualifiers.get("type", "tar.gz")
            target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

            self.get_logger().info(msg="Copying maven artifact:", indent_level=2)
            self.get_logger().info(msg=f"source: {repo_url}", indent_level=3)
            self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

            settings_dict = {
                "settings": {
                    "servers": {
                        "server": {
                            "id": "repoId",
                            "username": "${repo.login}",
                            "password": "${repo.pwd}",
                        }
                    }
                }
            }

            with NamedTemporaryFile(mode="w+", encoding="utf-8") as settings_file:
                settings_file.write(xmltodict.unparse(input_dict=settings_dict, pretty=True))
                settings_file.flush()

                password_list = []

                defines = {
                    "artifact": f"{artifact}:{extension}",
                    "outputDirectory": target_dir,
                    "destFileName": f"{purl.name}_{purl.version}.{extension}",
                    "remoteRepositories": f"repoId::::{repo_url}",
                }

                if creds is not None and isinstance(creds.password, SecretStr):
                    defines["repo.login"] = creds.username
                    defines["repo.pwd"] = creds.password.get_secret_value()
                    password_list = [creds.password.get_secret_value()]

                command = [
                    self.required_commands[0],
                    f"{_MAVEN_DEP_PLUGIN}:copy",
                    f"--settings={settings_file.name}",
                    *self.extra_opts,
                    *(["--debug"] if self.get_logger().is_verbose() else []),
                ]

                run_result = self._get_maven_component(command, password_list, **defines)
                error_msg = f"Failed to download maven artifact {artifact} type={extension}"

                try:
                    run_result.check_returncode()

                    if extension != "pom":
                        defines["artifact"] = str(defines["artifact"]).replace(f":{extension}", ":pom")
                        defines["destFileName"] = str(defines["destFileName"]).replace(f".{extension}", ".pom")

                        error_msg = f"Failed to download pom for maven artifact {artifact}"
                        run_result = self._get_maven_component(command, password_list, **defines)
                        run_result.check_returncode()
                except CalledProcessError:
                    self.get_logger().debug(msg=error_msg, indent_level=2)
                    return Result.retry(message=error_msg)

            self.set_collection_params(comp, repo_url, target_dir)

            return Result.success(return_obj=comp)



"""
Collector plugin for docker images
"""
from __future__ import annotations

import re
import urllib.parse

from pathlib import Path
from subprocess import CalledProcessError

from packageurl import PackageURL

from pydantic import SecretStr

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectDockerPlugin(SerialCollectorPlugin):
    """
    Collector plugin for docker images
    """

    supported_purl_types = ["docker", "oci"]
    required_commands = ["skopeo"]
    products: list[str] = ["docker/*", "oci/*"]
    system_repositories: list[str] = ["https://docker.io/"]
    process_timeout = 300

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        self.required_commands = (self.config or {}).get("skopeo_command", self.required_commands)

    def _get_image(self, url: str, purl: PackageURL) -> RepositoryUrl:
        """
        Return the image details for skopeo to process

        Args:
            url (str): Repository URL
            purl (PackageURL): Purl of component to operate on

        Returns:
            RepositoryUrl: Image information
        """
        # Determine if purl version contains SHA string, determine proper formatting for skopeo command
        match purl.version:
            case no_sha if re.search(r"^(sha256:)?[a-fA-F0-9]{12,64}$", no_sha) is None:
                image_name = f"{purl.name}:{purl.version}"
            case sha if sha.startswith("sha256:"):
                image_name = f"{purl.name}@{purl.version}"
            case _:
                image_name = f"{purl.name}@sha256:{purl.version}"

        if purl.type == "oci" and "repository_url" in purl.qualifiers:
            url = purl.qualifiers.get("repository_url", "")
            url = url.replace(purl.name, "")

        source_image = RepositoryUrl(url=url) / (purl.namespace or "") / urllib.parse.quote_plus(image_name)

        if source_image.scheme != "docker":
            source_image = RepositoryUrl(url="docker://" + re.sub(r"^(.*://)", "", str(source_image)))

        return source_image

    def _get_target_path(self, repo_url: str, purl: PackageURL) -> Path:
        """
        Get target path for image download

        Args:
            repo_url (str): Repository URL
            purl (PackageURL): Purl of component to operate on

        Returns:
            Path: Filesystem location for downloaded image
        """
        version = re.sub(pattern=r"^(sha256:)?([a-f0-9]{64})$", repl=r"sha256:\2", string=purl.version)

        version = urllib.parse.quote_plus(re.sub(r"^https?://", "", version))
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

        return target_dir / f"{purl.name}@{version}"

    def _inspect_image(self, source_image: RepositoryUrl, purl: PackageURL) -> Result:
        """
        Verify provided image tag digest matches PURL version

        Args:
            source_image (RepositoryUrl): Full image name
            purl (PackageURL): Purl of component to operate on

        Returns:
            Result: The result of the verification
        """
        image_prefix, _ = source_image.url.rsplit("/", maxsplit=1)
        image_id = f"{purl.name}:{purl.qualifiers.get('tag')}"

        command = [self.required_commands[0], "inspect", "--format", "{{.Digest}}", f"{image_prefix}/{image_id}"]
        inspect_command = self.run_command(command)

        try:
            inspect_command.check_returncode()
            sha_tag = inspect_command.stdout.decode().strip()
        except CalledProcessError:
            return Result.retry(message=f"Failed to get image digest for '{source_image}'")

        if sha_tag != purl.version:
            return Result.fail(
                message=f"Provided tag '{purl.qualifiers.get('tag')}' image digest does not match '{purl.version}'"
            )

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = hoppr.utils.get_package_url(comp.purl)
        source_image = self._get_image(url=repo_url, purl=purl)
        target_path = self._get_target_path(repo_url, purl)

        if purl.type == "oci" and "tag" in purl.qualifiers:
            if not (inspect_result := self._inspect_image(source_image, purl)).is_success():
                return inspect_result

        self.get_logger().info(msg=f"Copying {purl.type} image:", indent_level=2)
        self.get_logger().info(msg=f"source: {source_image}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_path}", indent_level=3)

        command = [self.required_commands[0], "copy"]
        password_list = []

        if creds is not None and isinstance(creds.password, SecretStr):
            password_list = [creds.password.get_secret_value()]
            command = [*command, "--src-creds", f"{creds.username}:{creds.password.get_secret_value()}"]

        if re.match("^http://", repo_url):
            command = [*command, "--src-tls-verify=false"]

        if self.get_logger().is_verbose():
            command = [*command, "--debug"]

        command = [*command, urllib.parse.unquote(str(source_image)), f"{purl.type}-archive:{target_path}"]
        copy_result = self.run_command(command, password_list)

        try:
            copy_result.check_returncode()
        except CalledProcessError as ex:
            msg = f"Skopeo failed to copy {purl.type} image to {target_path}, return_code={ex.returncode}"
            self.get_logger().debug(msg=msg, indent_level=2)

            self.get_logger().info(msg="Artifact collection failed, deleting file and retrying", indent_level=2)
            target_path.unlink(missing_ok=True)

            return Result.retry(message=msg)

        self.set_collection_params(comp, repo_url, target_path.parent)

        return Result.success(return_obj=comp)



"""
Hoppr Network utility functions
"""
from __future__ import annotations

import hashlib

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Final

import requests

from pydantic import SecretStr
from requests.auth import HTTPBasicAuth

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.utils import load_string

if TYPE_CHECKING:
    from os import PathLike

_BLOCK_SIZE: Final[int] = 65536


class HashAlgorithm(str, Enum):
    """Hash algorithms supported by builtin `hashlib` library."""

    BLAKE2B = "blake2b"
    BLAKE2S = "blake2s"
    MD5 = "md5"
    SHA1 = "sha1"
    SHA224 = "sha224"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"

    def __str__(self) -> str:
        return self.value


def get_file_hash(artifact: str | PathLike[str], algorithm: HashAlgorithm | str = HashAlgorithm.SHA1) -> str:
    """
    Compute hash of downloaded component.

    Args:
        artifact: Path to downloaded file
        algorithm: Hashing algorithm to use. Defaults to HashAlgorithm.SHA1.

    Returns:
        The computed hexidecimal hash digest.
    """
    artifact = Path(artifact)
    hash_obj = hashlib.new(name=str(algorithm))

    with artifact.open(mode="rb") as hash_bytes:
        while file_bytes := hash_bytes.read(_BLOCK_SIZE):
            hash_obj.update(file_bytes)

    return hash_obj.hexdigest().lower()


def load_url(url: str):
    """
    Load config content (either json or yml) from a url into a dict
    """
    creds = Credentials.find(url)

    response = None
    if creds is not None and isinstance(creds.password, SecretStr):
        authorization_headers = {
            "PRIVATE-TOKEN": creds.password.get_secret_value(),
            "Authorization": f"Bearer {creds.password.get_secret_value()}",
        }

        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
        response = requests.get(url, auth=basic_auth, headers=authorization_headers, timeout=60)
    else:
        response = requests.get(url, timeout=60)

    response.raise_for_status()
    valid_data = True
    try:
        if isinstance(response.content, bytes):
            return load_string(response.content.decode("utf-8"))
        if isinstance(response.content, str):
            return load_string(response.content)
        valid_data = False
    except HopprLoadDataError as parse_error:
        message = f"Unable to parse result from {url}."
        if response.url != url:
            message += f" Request was redirected to {response.url}. An auth issue might have occurred."
        raise HopprLoadDataError(message) from parse_error

    if not valid_data:
        raise HopprLoadDataError("Response type is not bytes or str")

    return None  # pragma: no cover


def download_file(
    url: str,
    dest: str,
    creds: CredentialRequiredService | None = None,
    proxies: dict[str, str] | None = None,
):
    """
    Download content from a url into a file
    """
    if creds is None:
        creds = Credentials.find(url)

    basic_auth = None
    if creds is not None and isinstance(creds.password, SecretStr):
        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

    response = requests.get(
        url=url,
        auth=basic_auth,
        allow_redirects=True,
        proxies=proxies,
        stream=True,
        verify=True,
        timeout=60,
    )

    if 200 <= response.status_code < 300:
        with open(dest, "wb") as out_file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    out_file.write(chunk)

    return response



"""
Collector plugin for pypi images
"""
from __future__ import annotations

import importlib.util
import re
import sys

from pydantic import SecretStr

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectPypiPlugin(SerialCollectorPlugin):
    """
    Collector plugin for pypi images
    """

    supported_purl_types = ["pypi"]
    required_commands = [sys.executable]
    products: list[str] = ["pypi/*"]
    system_repositories = ["https://pypi.org/simple"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.manifest_repos: list[str] = []
        self.password_list: list[str] = []
        self.base_command = [self.required_commands[0], "-m", "pip"]

        # Allow users to define their own pip command (e.g. "python3.6 -m pip")
        self.base_command = (self.config or {}).get("pip_command", f"{self.required_commands[0]} -m pip").split(" ")

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Copy a component to the local collection directory structure
        """
        if importlib.util.find_spec(name="pip") is None:
            return Result.fail(message="The pip package was not found. Please install and try again.")

        purl = hoppr.utils.get_package_url(comp.purl)

        source_url = RepositoryUrl(url=repo_url)
        if not re.match(pattern="^.*simple/?$", string=f"{source_url}"):
            source_url /= "simple"

        password_list = []

        if creds is not None and isinstance(creds.password, SecretStr):
            source_url = RepositoryUrl(
                url=source_url.url,
                username=creds.username,
                password=creds.password.get_secret_value(),
            )
            password_list = [creds.password.get_secret_value()]

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.name}_{purl.version}")

        self.get_logger().info(msg=f"Target directory: {target_dir}", indent_level=2)

        command = [
            *self.base_command,
            "download",
            "--no-deps",
            "--no-cache",
            "--timeout",
            str(self.process_timeout),
            "--index-url",
            f"{source_url}",
            "--dest",
            f"{target_dir}",
            f"{purl.name}=={purl.version}",
            *(["--verbose"] if self.get_logger().is_verbose() else []),
        ]

        base_error_msg = f"Failed to download {purl.name} version {purl.version}"

        collection_type = "binary-preferred"
        if self.config is not None and "type" in self.config:
            collection_type = str(self.config["type"]).lower()

        success_count = 0
        successes_needed = 2 if collection_type == "both-required" else 1
        match collection_type:
            case "binary" | "binary-only" | "binary-preferred" | "both-preferred" | "both-required":
                type_order = ["binary", "source"]
            case "source" | "source-only" | "source-preferred":
                type_order = ["source", "binary"]
            case _:
                return Result.fail(
                    f"{base_error_msg}. Invalid pypi collection type specified: {collection_type}", return_obj=comp
                )

        self.get_logger().info("Pypi collection type: %s", collection_type, indent_level=2)
        self.get_logger().info("Attempting %s collection", type_order[0], indent_level=2)

        run_result = self.run_command([*command, self._pypi_param_for_type(type_order[0])], password_list)

        if run_result.returncode == 0:
            success_count += 1

            if not collection_type.startswith("both"):
                self.get_logger().debug(
                    "Collection of %s successful, no %s collection needed", type_order[0], type_order[1], indent_level=2
                )
                self.set_collection_params(comp, repo_url, target_dir)

                return Result.success(return_obj=comp)
        elif not collection_type.endswith("preferred"):
            self.get_logger().debug(
                "Collection of %s failed, no %s collection needed", type_order[0], type_order[1], indent_level=2
            )

            return Result.retry(f"{base_error_msg}. Unable to collect {type_order[0]}", comp)

        self.get_logger().info("Attempting %s collection", type_order[1], indent_level=2)

        run_result = self.run_command([*command, self._pypi_param_for_type(type_order[1])], password_list)

        if run_result.returncode == 0:
            success_count += 1

        if success_count >= successes_needed:
            self.set_collection_params(comp, repo_url, target_dir)
            return self._success_result(collection_type, type_order, success_count, run_result.returncode, comp)

        return self._retry_result(collection_type, type_order, success_count, base_error_msg, comp)

    @staticmethod
    def _success_result(
        collection_type: str, type_order: list[str], success_count: int, last_rc: int, comp: Component
    ) -> Result:
        message = ""
        match collection_type:
            case "binary-preferred" | "source-preferred":
                message = f"Unable to download {type_order[0]}, {type_order[1]} collected"
            case "both-preferred" if success_count == 1 and last_rc == 0:
                message = f"Only able to download {type_order[1]}"
            case "both-preferred" if success_count == 1:
                message = f"Only able to download {type_order[0]}"

        return Result.success(message, return_obj=comp)

    @staticmethod
    def _retry_result(
        collection_type: str, type_order: list[str], success_count: int, base_error_msg: str, comp: Component
    ) -> Result:
        message = ""
        match collection_type:
            case "both-required" if success_count == 1:
                message = f"{base_error_msg}. Only able to download {type_order[0]}"
            case "binary-preferred" | "source-preferred" | "both-required" | "both-preferred":
                message = f"{base_error_msg}. Unable to download {type_order[0]} or {type_order[1]}."

        return Result.retry(message, return_obj=comp)

    @staticmethod
    def _pypi_param_for_type(collection_type: str) -> str:
        if collection_type == "source":
            return "--no-binary=:all:"

        return "--only-binary=:all:"



"""
Result models for `hopctl validate sbom`
"""
from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Any, Callable, Iterator, Mapping, TypeAlias

from pydantic import Field
from typing_extensions import Self, override

import hoppr.utils

from hoppr.models.base import HopprBaseModel
from hoppr.result import Result, ResultStatus

DictStrAny: TypeAlias = dict[str, Any]


def _encode_result(result: ValidateCheckResult) -> Mapping[str, str]:
    return {"status": result.status.name, "message": result.message}


class ValidateOutputFormat(str, Enum):
    """
    Output format for `hopctl validate sbom` when specifying an output file
    """

    JSON = "json"


class ValidateCheckResult(Result):
    """
    `Result` that can be deserialized from JSON
    """

    def __init__(self, status: ResultStatus | str, message: str = "", return_obj: Any | None = None):
        super().__init__(status=ResultStatus[str(status)], message=message, return_obj=return_obj)

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable]:
        yield cls.validate

    @classmethod
    def validate(cls, result: Self | DictStrAny) -> Self:
        """
        Validator to deserialize a result status JSON string into a `Result` object
        """
        if isinstance(result, dict):
            message: str = result.get("message", "")
            status: str = str(result.get("status"))

            result = cls(status=ResultStatus[status], message=message)

        return result

    @override
    def merge(self, other: Result):
        super().merge(other)

        # Remove duplicate messages from this result
        messages = self.message.split("\n")
        messages = hoppr.utils.dedup_list(messages)

        self.message = "\n".join(messages)


class ValidateBaseModel(HopprBaseModel, json_encoders={ValidateCheckResult: _encode_result}):
    """
    Base model for `hopctl validate sbom` results
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and all(
            getattr(self, name) == getattr(other, name) for name in self.__fields__
        )


class ValidateLicenseResult(ValidateBaseModel):
    """
    Results of `hopctl validate sbom` for an individual license
    """

    license_id: str
    expiration: ValidateCheckResult = Field(default=ValidateCheckResult.success(message="License Expiration"))
    required_fields: ValidateCheckResult = Field(default=ValidateCheckResult.success(message="Required License Fields"))


class ValidateComponentResult(ValidateBaseModel):
    """
    Results of `hopctl validate sbom` for an individual component
    """

    component_id: str
    license_results: list[ValidateLicenseResult] = Field(default=[])
    ntia_fields_result: ValidateCheckResult = Field(default=ValidateCheckResult.success(message="Minimum NTIA Fields"))
    result: ValidateCheckResult = Field(default=ValidateCheckResult.success())


class ValidateSbomResult(ValidateBaseModel):
    """
    Results of `hopctl validate sbom` for an SBOM file
    """

    sbom_file: Path
    component_results: list[ValidateComponentResult] = Field(default=[])
    license_results: list[ValidateLicenseResult] = Field(default=[])
    ntia_fields_result: ValidateCheckResult = Field(default=ValidateCheckResult.success(message="Minimum NTIA Fields"))
    result: ValidateCheckResult = Field(default=ValidateCheckResult.success())
    spec_version_result: ValidateCheckResult = Field(
        default=ValidateCheckResult.success(message="CycloneDX Specification Version")
    )
